#ifndef GLOBAL_H
#define GLOBAL_H

extern unsigned int samplestereo;
extern unsigned int samplesize;
extern unsigned int samplerate;

extern unsigned char OutputMap[8192];
extern unsigned char VolumeMap[65][256];
extern unsigned char MixBuffer[4096];

#endif
